import logging
import os

import requests

from .page import Page
from .project import Project
from .wiki import Wiki

logger = logging.getLogger(__name__)


class AzureDevOps:

    def __init__(self, organization, pat):
        self.organization = organization
        self._pat = pat
        self._auth = ("", self._pat)

    def get_projects(self, project_name=None):
        # https://docs.microsoft.com/en-us/rest/api/azure/devops/core/projects/list?view=azure-devops-rest-6.0
        url = f"https://dev.azure.com/{self.organization}/_apis/projects?api-version=6.0"
        logger.info(url)
        # https://docs.microsoft.com/en-us/azure/devops/organizations/accounts/use-personal-access-tokens-to-authenticate?view=azure-devops&tabs=preview-page
        response = requests.get(url, auth=self._auth)
        response.raise_for_status()
        projects = []
        for item in response.json()["value"]:
            _id = item["id"]
            name = item["name"]
            if name.strip().upper() == (project_name if project_name is not None else name).strip().upper():
                projects.append(Project(_id, name))

        for project in projects:
            project.wikis = self._get_wikis(project.id)
            for wiki in project.wikis:
                wiki.pages = self._get_pages(project.id, wiki.id)

        return projects

    def _get_wikis(self, project_id):
        # https://docs.microsoft.com/en-us/rest/api/azure/devops/wiki/wikis/list?view=azure-devops-rest-6.0
        url = f"https://dev.azure.com/{self.organization}/{project_id}/_apis/wiki/wikis?api-version=6.0"
        logger.info(url)
        response = requests.get(url, auth=self._auth)
        response.raise_for_status()
        return list(map(lambda item: Wiki(item["id"], item["name"]),
                        filter(lambda item: item["type"] == "projectWiki", response.json()["value"])))

    def _get_pages(self, project_id, wiki_id, continuation_token=None):
        # https://docs.microsoft.com/en-us/rest/api/azure/devops/wiki/pages%20batch/get?view=azure-devops-rest-6.1
        url = f"https://dev.azure.com/{self.organization}/{project_id}/_apis/wiki/wikis/{wiki_id}/pagesbatch" \
              f"?api-version=6.1-preview.1"
        logger.info(url)
        json = {"top": 10}
        if continuation_token is not None:
            json["continuationToken"] = continuation_token
        response = requests.post(url, auth=self._auth, json=json)
        response.raise_for_status()

        def create_page(item):
            page_id = item["id"]
            order, is_parent, content = self._get_page_details(project_id, wiki_id, page_id)
            return Page(page_id, item["path"], order, is_parent, content)

        pages = list(map(lambda item: create_page(item), response.json()["value"]))
        continuation_token = response.headers.get("x-ms-continuationtoken")
        if continuation_token is not None:
            pages.extend(self._get_pages(project_id, wiki_id, continuation_token))
        return pages

    def _get_page_details(self, project_id, wiki_id, page_id):
        # https://docs.microsoft.com/en-us/rest/api/azure/devops/wiki/pages/get%20page%20by%20id?view=azure-devops-rest-6.1
        url = f"https://dev.azure.com/{self.organization}/{project_id}/_apis/wiki/wikis/{wiki_id}/pages/{page_id}" \
              f"?api-version=6.1-preview.1"
        logger.info(url)
        params = {"includeContent": True, "recursionLevel": "none"}
        response = requests.get(url, auth=self._auth, params=params)
        response.raise_for_status()
        json_data = response.json()
        return int(json_data["order"]), json_data.get("isParentPage", False), json_data["content"]

    def download_attachment(self, project_id, wiki_id, attachment, directory):
        url = f"https://dev.azure.com/{self.organization}/{project_id}/_apis/git/repositories/{wiki_id}/Items" \
              f"?path={attachment}"
        logger.info(url)
        with requests.get(url, auth=self._auth, stream=True) as r:
            try:
                r.raise_for_status()
            except requests.exceptions.HTTPError as e:
                logger.error(f"Exception occurred downloading {url}", exc_info=True)
                if e.response.status_code != requests.codes.not_found:
                    raise
                return
            os.makedirs(directory, exist_ok=True)
            file_path = os.path.join(directory, attachment.split("/")[-1])
            with open(file_path, "wb") as f:
                for chunk in r.iter_content():
                    f.write(chunk)



import sys
import os


def is_debug_active():
    # https://stackoverflow.com/questions/38634988/check-if-program-runs-in-debug-mode
    gettrace = getattr(sys, 'gettrace', None)
    if gettrace is not None and gettrace():
        return True
    return False


def get_environment_variable(key):
    value = None
    try:
        value = os.environ[key]
    except KeyError:
        pass
    return value



import logging
import os
import re
import shutil
from datetime import datetime, timezone

logger = logging.getLogger(__name__)


class Hugo:

    def __init__(self, azure_devops):
        self._azure_devops = azure_devops

    def create_content(self, projects, site_directory):
        content_directory = os.path.join(site_directory, "content")
        self.__class__._empty_directory(content_directory)

        static_directory = os.path.join(site_directory, "static")
        self.__class__._empty_directory(static_directory)

        for project in projects:
            project_directory = os.path.join(content_directory, self.__class__._sanitize_path(project.name))

            for wiki in project.wikis:
                wiki_directory = os.path.join(project_directory, self.__class__._sanitize_path(wiki.name))

                for page in wiki.pages:
                    if not page.content:
                        continue

                    page_directories = list(map(
                        lambda item: self.__class__._sanitize_path(item), page.path.split("/")
                    ))

                    if page.is_parent:
                        page_directory = os.path.join(wiki_directory, *page_directories)
                        file_path = os.path.join(page_directory, "_index.md")
                    else:
                        page_directory = os.path.join(wiki_directory, *page_directories[:-1])
                        file_path = os.path.join(page_directory, f"{page_directories[-1]}.md")

                    self._extract_attachments(project.id, wiki.id, page.content, static_directory)
                    content = self.__class__._update_attachments(page.content, "/")

                    content = self.__class__._get_front_matter(page.title, page.is_parent) + content

                    self.__class__._create_file(file_path, content)

        self.__class__._create_index_page(content_directory, self._azure_devops.organization)
        for root, dirs, files in os.walk(content_directory):
            for _dir in dirs:
                self.__class__._create_index_page(os.path.join(root, _dir))

    @staticmethod
    def _get_front_matter(title, collapse=False):
        title = title.replace('"', '\\"')
        front_matter = f"""---
title: "{title}"
date: {datetime.now(timezone.utc).astimezone().isoformat(timespec="seconds")}
draft: true
"""
        if collapse:
            front_matter += "geekdocCollapseSection: true\n"
        front_matter += "---" + ("\n" * 2)
        return front_matter

    @classmethod
    def _create_index_page(cls, directory, title=None):
        index_path = os.path.join(directory, "_index.md")
        if os.path.exists(index_path):
            return
        logger.info(f"Creating {index_path}")
        with open(index_path, "x", encoding="utf-8") as f:
            title = title if title else os.path.basename(directory)
            content = cls._get_front_matter(title, True) + "{{< toc-tree >}}"
            f.write(content)

    @staticmethod
    def _empty_directory(directory):
        logger.info(f"Emptying {directory}")
        if not os.path.exists(directory):
            return
        for _dir in os.listdir(directory):
            path = os.path.join(directory, _dir)
            if os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.unlink(path)

    @staticmethod
    def _create_file(file_path, content):
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        logger.info(f"Creating {file_path}")
        with open(file_path, mode="x", encoding="utf-8") as f:
            f.write(content)

    def _extract_attachments(self, project_id, wiki_id, content, directory):
        pattern = r"\(\/?\.attachments\/.+?\)"
        for attachment in re.findall(pattern, content, re.MULTILINE):
            attachment = attachment[1:-1]
            if attachment.startswith("/"):
                attachment = attachment[1:]
            self._azure_devops.download_attachment(project_id, wiki_id, attachment, directory)

    @staticmethod
    def _update_attachments(content, location):
        pattern = r"\/?\.attachments\/"
        return re.sub(pattern, location, content)

    @staticmethod
    def _sanitize_path(path):
        # https://docs.microsoft.com/es-es/windows/win32/fileio/naming-a-file
        pattern = r'[<>:"/\\|?*]'
        return re.sub(pattern, "__", path)



# https://www.python.org/dev/peps/pep-0396/
# https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
# version as tuple for simple comparisons
VERSION = (1, 0, 0)
# string created from tuple to avoid inconsistency
__version__ = ".".join([str(x) for x in VERSION])



class Page:
    def __init__(self, _id, path, order, is_parent, content):
        self.id = _id
        self.path = path.lstrip("/")
        self.title = path.split("/")[-1]
        self.order = order
        self.is_parent = is_parent
        self.content = content



from datetime import timedelta
import logging
from timeit import default_timer as timer
import os
import sys

import jsonpickle

from .azure_devops import AzureDevOps
from .hugo import Hugo
from .utilities import is_debug_active, get_environment_variable


def main():
    logging.basicConfig(level=logging.INFO)
    start = timer()

    site_directory = sys.argv[1]
    if not os.path.exists(site_directory):
        logging.error(f"{site_directory} does not exist.")
        return

    project_name = None
    if len(sys.argv) == 3:
        project_name = sys.argv[2]

    organization = get_environment_variable("ORGANIZATION")
    if organization is None:
        logging.error("ORGANIZATION environment variable not found.")
        return

    pat = get_environment_variable("PAT")
    if pat is None:
        logging.error("PAT environment variable not found.")
        return

    azure_devops = AzureDevOps(organization, pat)

    projects = azure_devops.get_projects(project_name)

    if is_debug_active():
        json_data = jsonpickle.encode(projects)
        with open("data.json", mode="w", encoding="utf-8") as f:
            f.write(json_data)

    # with open("data.json", mode="r") as f:
    #     json_data = f.read()
    #     projects = jsonpickle.decode(json_data)

    hugo = Hugo(azure_devops)
    hugo.create_content(projects, site_directory)

    end = timer()
    logging.info(timedelta(seconds=end - start))


if __name__ == '__main__':
    main()



from setuptools import setup, find_packages
import toml

from src.ado2hugo import __version__


def get_long_description():
    with open("README.md") as file:
        return file.read()


def get_install_requires():
    data = toml.load("Pipfile")
    return [package + (version if version != "*" else "")
            for package, version in data["packages"].items()]


long_description = get_long_description()
packages = find_packages(where="src")
install_requires = get_install_requires()

setup(
    name="ado2hugo",
    version=__version__,
    author="panicoenlaxbox",
    author_email="panicoenlaxbox@gmail.com",
    description="Azure DevOps Wiki to Hugo",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=packages,
    package_dir={"": "src"},
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'ado2hugo = ado2hugo.main:main',
        ],
    },
)
